// Source: https://www.geeksforgeeks.org/sieve-of-eratosthenes/
// C++ program to print all primes smaller than or equal to
// n using Sieve of Eratosthenes

// Improved by Dan Strano of Unitary Fund, 2024.
// We can think of trial division as exact inverse of
// Sieve of Eratosthenes, with log space and log time.
// The modular division part is a costly atomic operation.
// It need only be carried out up the square root of the
// number under trial. Multiples of 2, 3, 5, 7, and 11 can
// be entirely skipped in loop enumeration.

#include "prime_generator.hpp"

#include <cmath>

#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace qimcifa {
std::vector<BigInteger> SieveOfEratosthenes(const BigInteger& n)
{
    std::vector<BigInteger> knownPrimes = { 2, 3, 5 };
    if (n < 2) {
        return std::vector<BigInteger>();
    }

    if (n < (knownPrimes.back() + 2)) {
        const auto highestPrimeIt = std::upper_bound(knownPrimes.begin(), knownPrimes.end(), n);
        return std::vector<BigInteger>(knownPrimes.begin(), highestPrimeIt);
    }

    knownPrimes.reserve(std::expint(log(n)) - std::expint(log(2)));

    // We are excluding multiples of the first few
    // small primes from outset. For multiples of
    // 2, 3, and 5 this reduces complexity to 4/15.
    const size_t cardinality = (size_t)backward5(n);

    // Create a boolean array "prime[0..cardinality]"
    // and initialize all entries it as true. Rather,
    // reverse the true/false meaning, so we can use
    // default initialization. A value in notPrime[i]
    // will finally be false only if i is a prime.
    size_t bitsPerWord = 64U;
    const size_t arrayWidth = (cardinality + bitsPerWord) / bitsPerWord;
    std::unique_ptr<uint64_t> uNotPrime(new uint64_t[arrayWidth]());
    std::unique_ptr<std::mutex> uNotPrimeMutex(new std::mutex[arrayWidth]());
    uint64_t* notPrime = uNotPrime.get();
    std::mutex* notPrimeMutex = uNotPrimeMutex.get();

    // We dispatch multiple marking asynchronously.
    // If we've already marked all primes up to x,
    // we're free to continue to up to x * x,
    // then we synchronize.
    BigInteger threadBoundary = 36U;

    // Get the remaining prime numbers.
    uint32_t wheel5 = (1U << 7U) | 1U;
    size_t o = 1U;
    for (;;) {
        o += GetWheel5Increment(wheel5);

        const BigInteger p = forward(o);
        if ((p * p) > n) {
            break;
        }

        if (threadBoundary < p) {
            dispatch.finish();
            threadBoundary *= threadBoundary;
        }

        const size_t q = (size_t)backward5(p);
        if (notPrime[q / bitsPerWord] == 1ULL << (q % bitsPerWord)) {
            continue;
        }

        knownPrimes.push_back(p);

        dispatch.dispatch([&n, p, &bitsPerWord, &notPrime, &notPrimeMutex]() {
            // We are skipping multiples of 2, 3, and 5
            // for space complexity, for 4/15 the bits.
            // More are skipped by the wheel for time.
            const BigInteger p2 = p << 1U;
            const BigInteger p4 = p << 2U;
            BigInteger i = p * p;

            // "p" already definitely not a multiple of 3.
            // Its remainder when divided by 3 can be 1 or 2.
            // If it is 2, we can do a "half iteration" of the
            // loop that would handle remainder of 1, and then
            // we can proceed with the 1 remainder loop.
            // This saves 2/3 of updates (or modulo).
            if ((p % 3U) == 2U) {
                const size_t q = (size_t)backward5(i);
                const size_t w = q / bitsPerWord;
                const uint64_t b = 1ULL << (q % bitsPerWord);
                if (true) {
                    std::lock_guard<std::mutex> lock(notPrimeMutex[w]);
                    notPrime[w] |= b;
                }
                i += p2;
                if (i > n) {
                    return false;
                }
            }

            for (;;) {
                if (i % 5U) {
                    const size_t q = (size_t)backward5(i);
                    const size_t w = q / bitsPerWord;
                    const uint64_t b = 1ULL << (q % bitsPerWord);
                    std::lock_guard<std::mutex> lock(notPrimeMutex[w]);
                    notPrime[w] |= b;
                }
                i += p4;
                if (i > n) {
                    return false;
                }

                if (i % 5U) {
                    const size_t q = (size_t)backward5(i);
                    const size_t w = q / bitsPerWord;
                    const uint64_t b = 1ULL << (q % bitsPerWord);
                    std::lock_guard<std::mutex> lock(notPrimeMutex[w]);
                    notPrime[w] |= b;
                }
                i += p2;
                if (i > n) {
                    return false;
                }
            }

            return false;
        });
    }

    dispatch.finish();
    notPrimeMutex = NULL;
    uNotPrimeMutex.reset();

    for (;;) {
        const BigInteger p = forward(o);
        if (p > n) {
            break;
        }

        o += GetWheel5Increment(wheel5);

        const size_t q = (size_t)backward5(p);
        if (notPrime[q / bitsPerWord] == 1ULL << (q % bitsPerWord)) {
            continue;
        }

        knownPrimes.push_back(p);
    }

    return knownPrimes;
}
} // namespace qimcifa

using namespace qimcifa;

PYBIND11_MODULE(eratosthenes, m) {
    m.doc() = "pybind11 plugin to generate prime numbers";
    m.def("sieve", &SieveOfEratosthenes, "Returns all primes up to the value of its argument (using Sieve of Eratosthenes)");
}
